'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var umiCore = require('@metaplex-foundation/umi-core');
var errors = require('./errors.cjs');
var numbers = require('./numbers.cjs');

class BeetSerializer {
  constructor(options = {}) {
    this.options = options;
  }
  tuple(items, description) {
    const itemDescriptions = items.map(item => item.description).join(', ');
    return {
      description: description ?? `tuple(${itemDescriptions})`,
      fixedSize: sumSerializerSizes(items.map(item => item.fixedSize)),
      maxSize: sumSerializerSizes(items.map(item => item.maxSize)),
      serialize: value => {
        if (value.length !== items.length) {
          throw new errors.BeetSerializerError(`Expected tuple to have ${items.length} items but got ${value.length}.`);
        }
        return umiCore.mergeBytes(items.map((item, index) => item.serialize(value[index])));
      },
      deserialize: (bytes, offset = 0) => {
        const values = [];
        items.forEach(serializer => {
          const [newValue, newOffset] = serializer.deserialize(bytes, offset);
          values.push(newValue);
          offset = newOffset;
        });
        return [values, offset];
      }
    };
  }
  vec(itemSerializer, prefix, description) {
    const prefixSeralizer = prefix ?? numbers.u32();
    return {
      description: description ?? `vec(${itemSerializer.description})`,
      fixedSize: null,
      maxSize: null,
      serialize: value => {
        const lengthBytes = prefixSeralizer.serialize(value.length);
        const itemBytes = value.map(item => itemSerializer.serialize(item));
        return umiCore.mergeBytes([lengthBytes, ...itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.length === 0) {
          return this.handleEmptyBuffer('vec', [], offset);
        }
        const values = [];
        const [length, newOffset] = prefixSeralizer.deserialize(bytes, offset);
        offset = newOffset;
        for (let i = 0; i < length; i += 1) {
          const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
          values.push(value);
          offset = newOffset;
        }
        return [values, offset];
      }
    };
  }
  array(itemSerializer, size, description) {
    function getSize(childSize) {
      if (size === 0) return 0;
      return childSize === null ? null : childSize * size;
    }
    return {
      description: description ?? `array(${itemSerializer.description}; ${size})`,
      fixedSize: getSize(itemSerializer.fixedSize),
      maxSize: getSize(itemSerializer.maxSize),
      serialize: value => {
        if (value.length !== size) {
          throw new errors.BeetSerializerError(`Expected array to have ${size} items but got ${value.length}.`);
        }
        return umiCore.mergeBytes(value.map(item => itemSerializer.serialize(item)));
      },
      deserialize: (bytes, offset = 0) => {
        const values = [];
        for (let i = 0; i < size; i += 1) {
          const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
          values.push(value);
          offset = newOffset;
        }
        return [values, offset];
      }
    };
  }
  map(keySerializer, valueSerializer, prefix, description) {
    const prefixSeralizer = prefix ?? numbers.u32();
    return {
      description: description ?? `map(${keySerializer.description}, ${valueSerializer.description})`,
      fixedSize: null,
      maxSize: null,
      serialize: map => {
        const lengthBytes = prefixSeralizer.serialize(map.size);
        const itemBytes = Array.from(map, ([key, value]) => umiCore.mergeBytes([keySerializer.serialize(key), valueSerializer.serialize(value)]));
        return umiCore.mergeBytes([lengthBytes, ...itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const map = new Map();
        if (bytes.length === 0) {
          return this.handleEmptyBuffer('map', map, offset);
        }
        const [length, newOffset] = prefixSeralizer.deserialize(bytes, offset);
        offset = newOffset;
        for (let i = 0; i < length; i += 1) {
          const [key, kOffset] = keySerializer.deserialize(bytes, offset);
          offset = kOffset;
          const [value, vOffset] = valueSerializer.deserialize(bytes, offset);
          offset = vOffset;
          map.set(key, value);
        }
        return [map, offset];
      }
    };
  }
  set(itemSerializer, prefix, description) {
    const prefixSeralizer = prefix ?? numbers.u32();
    return {
      description: description ?? `set(${itemSerializer.description})`,
      fixedSize: null,
      maxSize: null,
      serialize: set => {
        const lengthBytes = prefixSeralizer.serialize(set.size);
        const itemBytes = Array.from(set, value => itemSerializer.serialize(value));
        return umiCore.mergeBytes([lengthBytes, ...itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const set = new Set();
        if (bytes.length === 0) {
          return this.handleEmptyBuffer('set', set, offset);
        }
        const [length, newOffset] = prefixSeralizer.deserialize(bytes, offset);
        offset = newOffset;
        for (let i = 0; i < length; i += 1) {
          const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
          offset = newOffset;
          set.add(value);
        }
        return [set, offset];
      }
    };
  }
  option(itemSerializer, prefix, description) {
    const prefixSeralizer = prefix ?? numbers.u8();
    return {
      description: description ?? `option(${itemSerializer.description})`,
      fixedSize: itemSerializer.fixedSize === 0 ? prefixSeralizer.maxSize : null,
      maxSize: sumSerializerSizes([prefixSeralizer.maxSize, itemSerializer.maxSize]),
      serialize: option => {
        const prefixByte = prefixSeralizer.serialize(Number(umiCore.isSome(option)));
        const itemBytes = umiCore.isSome(option) ? itemSerializer.serialize(option.value) : new Uint8Array();
        return umiCore.mergeBytes([prefixByte, itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.length === 0) {
          return this.handleEmptyBuffer('option', umiCore.none(), offset);
        }
        const [isSome, prefixOffset] = prefixSeralizer.deserialize(bytes, offset);
        offset = prefixOffset;
        if (isSome === 0) {
          return [umiCore.none(), offset];
        }
        const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
        offset = newOffset;
        return [umiCore.some(value), offset];
      }
    };
  }
  fixedOption(itemSerializer, prefix, description) {
    const prefixSeralizer = prefix ?? numbers.u8();
    if (itemSerializer.fixedSize === null || prefixSeralizer.fixedSize === null) {
      throw new errors.BeetSerializerError('fixedOption can only be used with fixed size serializers');
    }
    return {
      description: description ?? `fixedOption(${itemSerializer.description})`,
      fixedSize: prefixSeralizer.fixedSize + itemSerializer.fixedSize,
      maxSize: prefixSeralizer.fixedSize + itemSerializer.fixedSize,
      serialize: option => {
        const fixedSize = itemSerializer.fixedSize;
        const prefixByte = prefixSeralizer.serialize(Number(umiCore.isSome(option)));
        const itemBytes = umiCore.isSome(option) ? itemSerializer.serialize(option.value).slice(0, fixedSize) : new Uint8Array(fixedSize).fill(0);
        return umiCore.mergeBytes([prefixByte, itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.length === 0) {
          return this.handleEmptyBuffer('fixedOption', umiCore.none(), offset);
        }
        const [isSome] = prefixSeralizer.deserialize(bytes, offset);
        offset += prefixSeralizer.fixedSize;
        const newOffset = offset + itemSerializer.fixedSize;
        if (isSome === 0) {
          return [umiCore.none(), newOffset];
        }
        const [value] = itemSerializer.deserialize(bytes, offset);
        return [umiCore.some(value), newOffset];
      }
    };
  }
  nullable(itemSerializer, prefix, description) {
    const prefixSeralizer = prefix ?? numbers.u8();
    return {
      description: description ?? `nullable(${itemSerializer.description})`,
      fixedSize: itemSerializer.fixedSize === 0 ? prefixSeralizer.maxSize : null,
      maxSize: sumSerializerSizes([prefixSeralizer.maxSize, itemSerializer.maxSize]),
      serialize: option => {
        const prefixByte = prefixSeralizer.serialize(Number(option !== null));
        const itemBytes = option !== null ? itemSerializer.serialize(option) : new Uint8Array();
        return umiCore.mergeBytes([prefixByte, itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.length === 0) {
          return this.handleEmptyBuffer('nullable', null, offset);
        }
        const [isSome, prefixOffset] = prefixSeralizer.deserialize(bytes, offset);
        offset = prefixOffset;
        if (isSome === 0) {
          return [null, offset];
        }
        const [value, newOffset] = itemSerializer.deserialize(bytes, offset);
        offset = newOffset;
        return [value, offset];
      }
    };
  }
  fixedNullable(itemSerializer, prefix, description) {
    const prefixSeralizer = prefix ?? numbers.u8();
    if (itemSerializer.fixedSize === null || prefixSeralizer.fixedSize === null) {
      throw new errors.BeetSerializerError('fixedNullable can only be used with fixed size serializers');
    }
    return {
      description: description ?? `fixedNullable(${itemSerializer.description})`,
      fixedSize: prefixSeralizer.fixedSize + itemSerializer.fixedSize,
      maxSize: prefixSeralizer.fixedSize + itemSerializer.fixedSize,
      serialize: option => {
        const fixedSize = itemSerializer.fixedSize;
        const prefixByte = prefixSeralizer.serialize(Number(option !== null));
        const itemBytes = option !== null ? itemSerializer.serialize(option).slice(0, fixedSize) : new Uint8Array(fixedSize).fill(0);
        return umiCore.mergeBytes([prefixByte, itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.length === 0) {
          return this.handleEmptyBuffer('fixedNullable', null, offset);
        }
        const [isSome] = prefixSeralizer.deserialize(bytes, offset);
        offset += prefixSeralizer.fixedSize;
        const newOffset = offset + itemSerializer.fixedSize;
        if (isSome === 0) {
          return [null, newOffset];
        }
        const [value] = itemSerializer.deserialize(bytes, offset);
        return [value, newOffset];
      }
    };
  }
  struct(fields, description) {
    const fieldDescriptions = fields.map(([name, serializer]) => `${String(name)}: ${serializer.description}`).join(', ');
    return {
      description: description ?? `struct(${fieldDescriptions})`,
      fixedSize: sumSerializerSizes(fields.map(([, field]) => field.fixedSize)),
      maxSize: sumSerializerSizes(fields.map(([, field]) => field.maxSize)),
      serialize: struct => {
        const fieldBytes = fields.map(([key, serializer]) => serializer.serialize(struct[key]));
        return umiCore.mergeBytes(fieldBytes);
      },
      deserialize: (bytes, offset = 0) => {
        const struct = {};
        fields.forEach(([key, serializer]) => {
          const [value, newOffset] = serializer.deserialize(bytes, offset);
          offset = newOffset;
          struct[key] = value;
        });
        return [struct, offset];
      }
    };
  }
  enum(constructor, description) {
    const enumValues = Object.values(constructor);
    const isNumericEnum = enumValues.some(v => typeof v === 'number');
    const valueDescriptions = enumValues.filter(v => typeof v === 'string').join(', ');
    function getVariantKeyValue(value) {
      if (typeof value === 'number') {
        return [enumValues[value], value];
      }
      const variantValue = constructor[value];
      if (typeof variantValue === 'number') {
        return [value, variantValue];
      }
      const indexOfValue = enumValues.indexOf(variantValue);
      if (indexOfValue >= 0) {
        return [variantValue, indexOfValue];
      }
      return [value, enumValues.indexOf(value)];
    }
    function checkVariantExists(variantKey) {
      if (!enumValues.includes(variantKey)) {
        throw new errors.BeetSerializerError(`Invalid enum variant. Got "${variantKey}", expected one of ` + `[${enumValues.join(', ')}]`);
      }
    }
    return {
      description: description ?? `enum(${valueDescriptions})`,
      fixedSize: 1,
      maxSize: 1,
      serialize: value => {
        const [variantKey, variantValue] = getVariantKeyValue(value);
        checkVariantExists(variantKey);
        return numbers.u8().serialize(variantValue);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.length === 0) {
          throw new errors.DeserializingEmptyBufferError('enum');
        }
        const [value, newOffset] = numbers.u8().deserialize(bytes, offset);
        offset = newOffset;
        const [variantKey, variantValue] = getVariantKeyValue(value);
        checkVariantExists(variantKey);
        return [isNumericEnum ? variantValue : variantKey, offset];
      }
    };
  }
  dataEnum(variants, prefix, description) {
    const prefixSeralizer = prefix ?? numbers.u8();
    const fieldDescriptions = variants.map(([name, serializer]) => `${String(name)}${serializer ? `: ${serializer.description}` : ''}`).join(', ');
    const allVariantHaveTheSameFixedSize = variants.every((one, i, all) => one[1].fixedSize === all[0][1].fixedSize);
    const fixedVariantSize = allVariantHaveTheSameFixedSize ? variants[0][1].fixedSize : null;
    const maxVariantSize = maxSerializerSizes(variants.map(([, field]) => field.maxSize));
    return {
      description: description ?? `dataEnum(${fieldDescriptions})`,
      fixedSize: variants.length === 0 ? prefixSeralizer.fixedSize : sumSerializerSizes([prefixSeralizer.fixedSize, fixedVariantSize]),
      maxSize: variants.length === 0 ? prefixSeralizer.maxSize : sumSerializerSizes([prefixSeralizer.maxSize, maxVariantSize]),
      serialize: variant => {
        const discriminator = variants.findIndex(([key]) => variant.__kind === key);
        if (discriminator < 0) {
          throw new errors.BeetSerializerError(`Invalid data enum variant. Got "${variant.__kind}", expected one of ` + `[${variants.map(([key]) => key).join(', ')}]`);
        }
        const variantPrefix = prefixSeralizer.serialize(discriminator);
        const variantSerializer = variants[discriminator][1];
        const variantBytes = variantSerializer.serialize(variant);
        return umiCore.mergeBytes([variantPrefix, variantBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.length === 0) {
          throw new errors.DeserializingEmptyBufferError('dataEnum');
        }
        const [discriminator, dOffset] = prefixSeralizer.deserialize(bytes, offset);
        offset = dOffset;
        const variantField = variants[Number(discriminator)] ?? null;
        if (!variantField) {
          throw new errors.BeetSerializerError(`Data enum index "${discriminator}" is out of range. ` + `Index should be between 0 and ${variants.length - 1}.`);
        }
        const [variant, vOffset] = variantField[1].deserialize(bytes, offset);
        offset = vOffset;
        return [{
          __kind: variantField[0],
          ...(variant ?? {})
        }, offset];
      }
    };
  }
  fixed(bytes, child, description) {
    return fixed(bytes, child, description);
  }
  string(prefix, content, description) {
    const prefixSerializer = prefix ?? numbers.u32();
    const contentSerializer = content ?? umiCore.utf8;
    return {
      description: description ?? `string(${prefixSerializer.description}, ${contentSerializer.description})`,
      fixedSize: null,
      maxSize: null,
      serialize: value => {
        const contentBytes = contentSerializer.serialize(value);
        const lengthBytes = prefixSerializer.serialize(contentBytes.length);
        return umiCore.mergeBytes([lengthBytes, contentBytes]);
      },
      deserialize: (buffer, offset = 0) => {
        if (buffer.length === 0) {
          throw new errors.DeserializingEmptyBufferError('string');
        }
        const [length, lengthOffset] = prefixSerializer.deserialize(buffer, offset);
        offset = lengthOffset;
        const contentBuffer = buffer.slice(offset, offset + Number(length));
        const [value, contentOffset] = contentSerializer.deserialize(contentBuffer);
        offset += contentOffset;
        return [value, offset];
      }
    };
  }
  fixedString(bytes, content, description) {
    const contentSerializer = content ?? umiCore.utf8;
    return fixed(bytes, contentSerializer, description ?? `fixedString(${bytes}, ${contentSerializer.description})`);
  }
  bool(size, description) {
    const serializer = size ?? numbers.u8();
    if (serializer.fixedSize === null) {
      throw new errors.BeetSerializerError('Serializer [bool] requires a fixed size.');
    }
    return {
      description: description ?? `bool(${serializer.description})`,
      fixedSize: serializer.fixedSize,
      maxSize: serializer.fixedSize,
      serialize: value => serializer.serialize(value ? 1 : 0),
      deserialize: (bytes, offset = 0) => {
        if (bytes.length === 0) {
          throw new errors.DeserializingEmptyBufferError('bool');
        }
        const [value, vOffset] = serializer.deserialize(bytes, offset);
        return [value === 1, vOffset];
      }
    };
  }
  get unit() {
    return {
      description: 'unit',
      fixedSize: 0,
      maxSize: 0,
      serialize: () => new Uint8Array(),
      deserialize: (_bytes, offset = 0) => [undefined, offset]
    };
  }
  get u8() {
    return numbers.u8();
  }
  get u16() {
    return numbers.u16();
  }
  get u32() {
    return numbers.u32();
  }
  get u64() {
    return numbers.u64();
  }
  get u128() {
    return numbers.u128();
  }
  get i8() {
    return numbers.i8();
  }
  get i16() {
    return numbers.i16();
  }
  get i32() {
    return numbers.i32();
  }
  get i64() {
    return numbers.i64();
  }
  get i128() {
    return numbers.i128();
  }
  get f32() {
    return {
      description: 'f32 [not supported]',
      fixedSize: 4,
      maxSize: 4,
      serialize: () => {
        throw new errors.OperationNotSupportedError('f32');
      },
      deserialize: () => {
        throw new errors.OperationNotSupportedError('f32');
      }
    };
  }
  get f64() {
    return {
      description: 'f64 [not supported]',
      fixedSize: 8,
      maxSize: 8,
      serialize: () => {
        throw new errors.OperationNotSupportedError('f64');
      },
      deserialize: () => {
        throw new errors.OperationNotSupportedError('f64');
      }
    };
  }
  get bytes() {
    return {
      description: 'bytes',
      fixedSize: null,
      maxSize: null,
      serialize: value => new Uint8Array(value),
      deserialize: (bytes, offset = 0) => [new Uint8Array(bytes), offset + bytes.length]
    };
  }
  get publicKey() {
    return {
      description: 'publicKey',
      fixedSize: 32,
      maxSize: 32,
      serialize: value => umiCore.publicKey(value).bytes,
      deserialize: (bytes, offset = 0) => {
        if (bytes.length === 0) {
          throw new errors.DeserializingEmptyBufferError('publicKey');
        }
        const pubkeyBytes = bytes.slice(offset, offset + 32);
        if (pubkeyBytes.length < 32) {
          throw new errors.BeetSerializerError(`Serializer [publicKey] expected 32 bytes, got ${pubkeyBytes.length}.`);
        }
        return [umiCore.publicKey(pubkeyBytes), offset + 32];
      }
    };
  }
  handleEmptyBuffer(serializer, defaultValue, offset) {
    if (!(this.options.tolerateEmptyBuffers ?? true)) {
      throw new errors.DeserializingEmptyBufferError(serializer);
    }
    return [defaultValue, offset];
  }
}
function fixed(bytes, child, description) {
  return {
    description: description ?? `fixed(${bytes}, ${child.description})`,
    fixedSize: bytes,
    maxSize: bytes,
    serialize: value => {
      const buffer = new Uint8Array(bytes).fill(0);
      buffer.set(child.serialize(value).slice(0, bytes));
      return buffer;
    },
    deserialize: (buffer, offset = 0) => {
      buffer = buffer.slice(offset, offset + bytes);
      if (buffer.length < bytes) {
        throw new errors.BeetSerializerError(`Serializer [fixed] expected ${bytes} bytes, got ${buffer.length}.`);
      }
      const [value] = child.deserialize(buffer, offset);
      return [value, offset + bytes];
    }
  };
}
function sumSerializerSizes(sizes) {
  return sizes.reduce((all, size) => all === null || size === null ? null : all + size, 0);
}
function maxSerializerSizes(sizes) {
  return sizes.reduce((all, size) => all === null || size === null ? null : Math.max(all, size), 0);
}

exports.BeetSerializer = BeetSerializer;
//# sourceMappingURL=BeetSerializer.cjs.map
